import os

from ..helper import add_arg_group


def mixin_hub_docker_login_parser(parser):
    gp = add_arg_group(parser, title='Docker login')

    gp.add_argument('--username', type=str, help='the Docker registry username',
                    default=os.environ.get('JINAHUB_USERNAME', ''))
    # _gp = parser.add_mutually_exclusive_group()
    # _gp.add_argument('--password-stdin', type=argparse.FileType('r'),
    #                  default=(sys.stdin if sys.stdin.isatty() else None),
    #                  help='take the password from stdin')
    gp.add_argument('--password', type=str, help='the plaintext password',
                    default=os.environ.get('JINAHUB_PASSWORD', ''))
    gp.add_argument('--registry', type=str, default='https://index.docker.io/v1/',
                    help='the URL to the Docker registry, e.g. https://index.docker.io/v1/')
    gp.add_argument('--repository', type=str, default='jinahub',
                    help='the Docker repository name, change this when pushing image to a personal repository')



import os
import shutil
from typing import Iterator, Optional

from ..flow import Flow
from ..helper import colored
from ..logging import default_logger as logger


class FlowRunner:
    """Module to define and run a flow."""

    def __init__(
        self,
        flow_yaml: str,
        documents: Iterator,
        request_size: int,
        task: str,  # this can be only index or search as it is used to call the flow API
        callback: Optional = None,
        overwrite_workspace: bool = False,
    ):
        """
        :param flow_yaml: path to flow yaml
        :param documents: iterator with list or generator for getting the documents
        :param request_size: request size used in the flow
        :param task: task of the flow which can be `index` or `search`
        :param callback: callback to be passed to the flow's `on_done`
        :param overwrite_workspace: overwrite workspace created by the flow
        """
        self.flow_yaml = flow_yaml
        # TODO: Make changes for working with doc generator (Pratik, before v1.0)
        self.documents = documents if type(documents) == list else list(documents)
        self.request_size = request_size
        if task in ('index', 'search'):
            self.task = task
        else:
            raise ValueError('task can be either of index or search')
        self.callback = callback
        self.overwrite_workspace = overwrite_workspace

    def _setup_workspace(self, workspace):
        if os.path.exists(workspace):
            if self.overwrite_workspace:
                shutil.rmtree(workspace)
                logger.warning(colored('Existing workspace deleted', 'red'))
                logger.warning(colored('WORKSPACE: ' + str(workspace), 'red'))
                logger.warning(
                    colored('change overwrite_workspace to change this', 'red')
                )
            else:
                logger.warning(
                    colored(
                        f'Workspace {workspace} already exists. Please set ``overwrite_workspace=True`` for replacing it.',
                        'red',
                    )
                )

        os.makedirs(workspace, exist_ok=True)

    def run(
        self,
        trial_parameters: dict,
        workspace: str = 'workspace',
        **kwargs,
    ):
        """[summary]

        :param trial_parameters: flow env variable values
        :param workspace: directory to be used for artifacts generated
        """

        self._setup_workspace(workspace)
        with Flow.load_config(self.flow_yaml, context=trial_parameters) as f:
            getattr(f, self.task)(
                self.documents,
                request_size=self.request_size,
                on_done=self.callback,
                **kwargs,
            )


class MultiFlowRunner:
    """Chain and run multiple flows"""

    def __init__(self, *flows):
        """
        :param flows: flows to be executed in sequence
        """
        self.flows = flows

    def run(
        self,
        trial_parameters: dict,
        workspace: str = 'workspace',
        **kwargs,
    ):
        """
        :param trial_parameters: parameters to be used as environment variables
        :param workspace: directory to be used for the flows
        """
        for flow in self.flows:
            flow.run(trial_parameters, workspace, **kwargs)



import inspect
from typing import Dict, List

from .. import BaseExecutor
from ...helper import typename


class BaseSegmenter(BaseExecutor):
    """:class:`BaseSegmenter` works on doc-level,
        it chunks Documents into set of Chunks """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.required_keys = {k for k in inspect.getfullargspec(self.segment).args if k != 'self'}
        if not self.required_keys:
            self.logger.warning(f'{typename(self)} works on keys, but no keys are specified')

    def segment(self, *args, **kwargs) -> List[Dict]:
        """The apply function of this executor.
        Unlike :class:`BaseCrafter`, the :func:`craft` here works on doc-level info and the output is defined on
        chunk-level. Therefore the name of the arguments should be always valid keys defined
        in the doc-level protobuf whereas the output dict keys should always be valid keys defined in the chunk-level
        protobuf.
        :return: a list of chunks-level info represented by a dict
        """
        raise NotImplementedError



import shutil
import uuid
from collections.abc import MutableMapping
from datetime import datetime
from typing import Dict, Any, Union

from jina.helper import colored
from jina.logging import JinaLogger
from .. import jinad_args


class BaseStore(MutableMapping):

    def __init__(self):
        self._items = {}  # type: Dict['uuid.UUID', Dict[str, Any]]
        self._logger = JinaLogger(self.__class__.__name__, **vars(jinad_args))
        self._init_stats()

    def _init_stats(self):
        """Initialize the stats """
        self._time_created = datetime.now()
        self._time_updated = self._time_created
        self._num_add = 0
        self._num_del = 0

    def add(self, *args, **kwargs) -> 'uuid.UUID':
        """Add a new element to the store. This method needs to be overridden by the subclass"""
        raise NotImplementedError

    def __iter__(self):
        return iter(self._items)

    def __len__(self):
        return len(self._items)

    def __getitem__(self, key: Union['uuid.UUID', str]):
        if isinstance(key, str):
            key = uuid.UUID(key)
        return self._items[key]

    def __delitem__(self, key: Union['uuid.UUID', str]):
        """ Release a Pea/Pod/Flow object from the store """
        if isinstance(key, str):
            key = uuid.UUID(key)

        if key in self._items:
            v = self._items[key]
            if 'object' in v and hasattr(v['object'], 'close'):
                v['object'].close()
            if v.get('workdir', None):
                shutil.rmtree(v['workdir'])
            self._items.pop(key)
            self._time_updated = datetime.now()
            self._logger.success(f'{colored(str(key), "cyan")} is released from the store.')
            self._num_del += 1
        else:
            raise KeyError(f'{colored(str(key), "cyan")} not found in store.')

    def clear(self) -> None:
        keys = list(self._items.keys())
        for k in keys:
            self.pop(k)

    def reset(self) -> None:
        """Calling :meth:`clear` and reset all stats """
        self.clear()
        self._init_stats()

    def __setitem__(self, key: 'uuid.UUID', value: Dict) -> None:
        self._items[key] = value
        t = datetime.now()
        value.update({'time_created': t})
        self._time_updated = t
        self._logger.success(f'{colored(str(key), "cyan")} is added')
        self._num_add += 1

    @property
    def status(self) -> Dict:
        """Return the status of this store as a dict"""
        return {
            'size': len(self._items),
            'time_created': self._time_created,
            'time_updated': self._time_updated,
            'num_add': self._num_add,
            'num_del': self._num_del,
            'items': self._items
        }



from typing import List, Any, Union

import numpy as np

from . import BaseExecutableDriver
from ..helper import typename
from ..types.document import Document

if False:
    from ..types.sets import DocumentSet


class BasePredictDriver(BaseExecutableDriver):
    """Drivers inherited from this Driver will bind :meth:`predict` by default """

    def __init__(self, executor: str = None, method: str = 'predict', *args, **kwargs):
        super().__init__(executor, method, *args, **kwargs)


class BaseLabelPredictDriver(BasePredictDriver):

    def __init__(self, output_tag: str = 'prediction', *args, **kwargs):
        """

        :param output_tag: output label will be written to ``doc.tags``
        :param args:
        :param kwargs:
        """
        super().__init__(*args, **kwargs)
        self.output_tag = output_tag

    def _apply_all(
            self,
            docs: 'DocumentSet',
            context_doc: 'Document',
            field: str,
            *args,
            **kwargs,
    ) -> None:
        embed_vecs, docs_pts, bad_docs = docs.all_embeddings

        if bad_docs:
            self.runtime.logger.warning(f'these bad docs can not be added: {bad_docs}')

        if docs_pts:
            prediction = self.exec_fn(embed_vecs)
            labels = self.prediction2label(prediction)  # type: List[Union[str, List[str]]]
            for doc, label in zip(docs_pts, labels):
                doc.tags[self.output_tag] = label

    def prediction2label(self, prediction: 'np.ndarray') -> List[Any]:
        """ Converting ndarray prediction into list of readable labels

        .. note::
            ``len(output)`` should be the same as ``prediction.shape[0]``

        :param prediction: the float/int numpy ndarray given by :class:`BaseClassifier`
        :return: the readable label to be stored.
        """
        raise NotImplementedError


class BinaryPredictDriver(BaseLabelPredictDriver):
    """ Converts binary prediction into string label.

    This is often used with binary classifier.
    """

    def __init__(self, one_label: str = 'yes', zero_label: str = 'no', *args, **kwargs):
        """

        :param one_label: label when prediction is one
        :param zero_label: label when prediction is zero
        :param args:
        :param kwargs:
        """
        super().__init__(*args, **kwargs)
        self.one_label = one_label
        self.zero_label = zero_label

    def prediction2label(self, prediction: 'np.ndarray') -> List[str]:
        """

        :param prediction: a (B,) or (B, 1) zero one array
        :return:
        """
        p = np.squeeze(prediction)
        if p.ndim > 1:
            raise ValueError(f'{typename(self)} expects prediction has ndim=1, but receiving ndim={p.ndim}')

        return [self.one_label if v else self.zero_label for v in p.astype(bool)]


class OneHotPredictDriver(BaseLabelPredictDriver):
    """ Mapping prediction to one of the given labels

    Expect prediction to be 2dim array, zero-one valued. Each row corresponds to
    a sample, each column corresponds to a label. Each row can have only one 1.

    This is often used with multi-class classifier.
    """

    def __init__(self, labels: List[str], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.labels = labels

    def validate_labels(self, prediction: 'np.ndarray'):
        if prediction.ndim != 2:
            raise ValueError(f'{typename(self)} expects prediction has ndim=2, but receiving {prediction.ndim}')
        if prediction.shape[1] != len(self.labels):
            raise ValueError(
                f'{typename(self)} expects prediction.shape[1]==len(self.labels), but receiving {prediction.shape}')

    def prediction2label(self, prediction: 'np.ndarray') -> List[str]:
        """

        :param prediction: a (B, C) array where C is the number of classes, only one element can be one
        :return:
        """
        self.validate_labels(prediction)
        p = np.argmax(prediction, axis=1)
        return [self.labels[v] for v in p]


class MultiLabelPredictDriver(OneHotPredictDriver):
    """ Mapping prediction to a list of labels

    Expect prediction to be 2dim array, zero-one valued. Each row corresponds to
    a sample, each column corresponds to a label. Each row can have only multiple 1s.

    This is often used with multi-label classifier, where each instance can have multiple labels
    """

    def prediction2label(self, prediction: 'np.ndarray') -> List[List[str]]:
        self.validate_labels(prediction)
        return [[self.labels[int(pp)] for pp in p.nonzero()[0]] for p in prediction]


class Prediction2DocBlobDriver(BasePredictDriver):
    """ Write the prediction result directly into ``document.blob``.

    .. warning::

        This will erase the content in ``document.text`` and ``document.buffer``.
    """

    def _apply_all(
            self,
            docs: 'DocumentSet',
            context_doc: 'Document',
            field: str,
            *args,
            **kwargs,
    ) -> None:
        embed_vecs, docs_pts, bad_docs = docs.all_embeddings

        if bad_docs:
            self.runtime.logger.warning(f'these bad docs can not be added: {bad_docs}')

        if docs_pts:
            prediction = self.exec_fn(embed_vecs)
            for doc, pred in zip(docs_pts, prediction):
                doc.blob = pred



import argparse

from ...helper import add_arg_group, _SHOW_ALL_ARGS
from .... import __default_host__
from ....enums import SkipOnErrorType, SocketType
from ....helper import random_port


def mixin_zed_runtime_parser(parser):
    """Mixing in arguments required by :class:`ZEDRuntime` into the given parser."""

    gp = add_arg_group(parser, title='ZEDRuntime')

    gp.add_argument('--uses', type=str, default='_pass',
                    help='the config of the executor, it could be '
                         '> a YAML file path, '
                         '> an executor\'s class name, '
                         '> a docker image (must start with docker://)'
                         '> one of "_pass", "_logforward" '
                         '> the raw content of YAML config (must starts with "!")')
    gp.add_argument('--py-modules', type=str, nargs='*', metavar='PATH',
                    help='the customized python modules need to be imported before loading the'
                         ' executor')

    gp.add_argument('--port-in', type=int, default=random_port(),
                    help='port for input data, default a random port between [49152, 65535]')
    gp.add_argument('--port-out', type=int, default=random_port(),
                    help='port for output data, default a random port between [49152, 65535]')
    gp.add_argument('--host-in', type=str, default=__default_host__,
                    help=f'host address for input, by default it is {__default_host__}')
    gp.add_argument('--host-out', type=str, default=__default_host__,
                    help=f'host address for output, by default it is {__default_host__}')
    gp.add_argument('--socket-in', type=SocketType.from_string, choices=list(SocketType),
                    default=SocketType.PULL_BIND,
                    help='socket type for input port')
    gp.add_argument('--socket-out', type=SocketType.from_string, choices=list(SocketType),
                    default=SocketType.PUSH_BIND,
                    help='socket type for output port')
    gp.add_argument('--max-socket-retries', type=int, default=3,
                    help='max number of retries when socket is conflict')

    gp.add_argument('--dump-interval', type=int, default=240,
                    help='serialize the model in the pod every n seconds if model changes. '
                         '-1 means --read-only. ')
    gp.add_argument('--read-only', action='store_true', default=False,
                    help='do not allow the pod to modify the model, '
                         'dump_interval will be ignored')
    gp.add_argument('--separated-workspace', action='store_true', default=False,
                    help='the data and config files are separated for each pea in this pod, '
                         'only effective when BasePod\'s `parallel` > 1')

    gp.add_argument('--memory-hwm', type=int, default=-1,
                    help='memory high watermark of this pod in Gigabytes, pod will restart when this is reached. '
                         '-1 means no restriction')

    gp.add_argument('--skip-on-error', type=SkipOnErrorType.from_string, choices=list(SkipOnErrorType),
                    default=SkipOnErrorType.NONE,
                    help='skip strategy on error message.')

    gp.add_argument('--num-part', type=int, default=0,
                    help='the number of messages expected from upstream, 0 and 1 means single part'
                    if _SHOW_ALL_ARGS else argparse.SUPPRESS)




from . import Request
from ...proto import jina_pb2

__all__ = ['TrainDryRunRequest', 'IndexDryRunRequest', 'SearchDryRunRequest', 'ControlDryRunRequest', 'DryRunRequest']


class DryRunRequest(Request):
    """Base empty request for dry run"""


class TrainDryRunRequest(DryRunRequest):
    """Empty train request for dry run"""

    def __init__(self):
        super().__init__()
        self.as_pb_object.train.CopyFrom(jina_pb2.RequestProto.TrainRequestProto())


class IndexDryRunRequest(DryRunRequest):
    """Empty index request for dry run"""

    def __init__(self):
        super().__init__()
        self.as_pb_object.index.CopyFrom(jina_pb2.RequestProto.IndexRequestProto())


class SearchDryRunRequest(DryRunRequest):
    """Empty search request for dry run"""

    def __init__(self):
        super().__init__()
        self.as_pb_object.search.CopyFrom(jina_pb2.RequestProto.SearchRequestProto())


class ControlDryRunRequest(DryRunRequest):
    """Empty control request for dry run"""

    def __init__(self):
        super().__init__()
        self.as_pb_object.control.CopyFrom(jina_pb2.RequestProto.ControlRequestProto())



from typing import Dict

import numpy as np
from jina.executors.crafters import BaseCrafter


class ArrayBytesReader(BaseCrafter):
    """
    :class:`ArrayBytesReader` converts a byte stream into a numpy array and save to the Document.
        The size of the vectors is provided in the constructor so that the numpy array can be interpreted properly
    """

    def __init__(self, as_type: str = 'float32', *args, **kwargs):
        """
        :param as_type: type of number
        """
        super().__init__(*args, **kwargs)
        self.as_type = as_type

    def craft(self, buffer: bytes, *args, **kwargs) -> Dict:
        """
        Split string into numbers and convert to numpy array

        :param buffer: the bytes representing the array
        :return: a chunk dict with the numpy array
        """
        _array = np.frombuffer(buffer, self.as_type)
        return dict(weight=1., blob=_array)



from typing import Optional, Sequence, Union

import inspect

from ..jaml import JAML


class OptimizationParameter:
    def __init__(
        self,
        parameter_name: str,
        executor_name: Optional[str] = None,
        prefix: str = 'JINA',
        env_var: Optional[str] = None,
    ):
        if env_var is None:
            self.env_var = f'{prefix}_{executor_name}_{parameter_name}'.upper()
        else:
            self.env_var = env_var
        self.parameter_name = parameter_name

    @classmethod
    def to_yaml(cls, representer, data):
        """Required by :mod:`pyyaml` """
        tmp = data._dump_instance_to_yaml(data)
        representer.sort_base_mapping_type_on_output = False
        return representer.represent_mapping('!' + cls.__name__, tmp)

    @staticmethod
    def _dump_instance_to_yaml(instance):

        attributes = inspect.getmembers(instance, lambda a: not (inspect.isroutine(a)))
        return {
            a[0]: a[1]
            for a in attributes
            if not (a[0].startswith('__') and a[0].endswith('__'))
        }

    @classmethod
    def from_yaml(cls, constructor, node):
        """Required by :mod:`pyyaml` """
        return cls._get_instance_from_yaml(constructor, node)

    @classmethod
    def _get_instance_from_yaml(cls, constructor, node):
        data = constructor.construct_mapping(node, deep=True)
        return cls(**data)


class IntegerParameter(OptimizationParameter):
    def __init__(
        self,
        low: int,
        high: int,
        step_size: int = 1,
        log: bool = False,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.low = low
        self.high = high
        self.step_size = step_size
        # The step != 1 and log arguments cannot be used at the same time.
        # To set the log argument to True, set the step argument to 1.
        self.log = log
        self.optuna_method = 'suggest_int'

    def to_optuna_args(self):
        return {
            'name': self.env_var,
            'low': self.low,
            'high': self.high,
            'step': self.step_size,
            'log': self.log,
        }


JAML.register(IntegerParameter)


class FloatParameter(OptimizationParameter):
    def __init__(
        self,
        low: float,
        high: float,
        step_size: Optional[float] = None,
        log: bool = False,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.low = low
        self.high = high
        self.step_size = step_size
        self.log = log
        self.optuna_method = 'suggest_float'

    def to_optuna_args(self):
        return {
            'name': self.env_var,
            'low': self.low,
            'high': self.high,
            'step': self.step_size,
            'log': self.log,
        }


JAML.register(FloatParameter)


class UniformParameter(OptimizationParameter):
    def __init__(self, low: float, high: float, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.low = low
        self.high = high
        self.optuna_method = 'suggest_uniform'

    def to_optuna_args(self):
        return {
            'name': self.env_var,
            'low': self.low,
            'high': self.high,
        }


JAML.register(UniformParameter)


class LogUniformParameter(OptimizationParameter):
    def __init__(self, low: float, high: float, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.low = low
        self.high = high
        self.optuna_method = 'suggest_loguniform'

    def to_optuna_args(self):
        return {
            'name': self.env_var,
            'low': self.low,
            'high': self.high,
        }


JAML.register(LogUniformParameter)


class CategoricalParameter(OptimizationParameter):
    def __init__(
        self, choices: Sequence[Union[None, bool, int, float, str]], *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.choices = choices
        self.optuna_method = 'suggest_categorical'

    def to_optuna_args(self):
        return {'name': self.env_var, 'choices': self.choices}


JAML.register(CategoricalParameter)


class DiscreteUniformParameter(OptimizationParameter):
    def __init__(self, low: float, high: float, q: float, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.low = low
        self.high = high
        self.q = q
        self.optuna_method = 'suggest_discrete_uniform'

    def to_optuna_args(self):
        return {
            'name': self.env_var,
            'low': self.low,
            'high': self.high,
            'q': self.q,
        }


JAML.register(DiscreteUniformParameter)


def load_optimization_parameters(filepath: str):
    JAML.register(IntegerParameter)
    JAML.register(FloatParameter)
    JAML.register(UniformParameter)
    JAML.register(LogUniformParameter)
    JAML.register(CategoricalParameter)
    JAML.register(DiscreteUniformParameter)
    with open(filepath, encoding='utf8') as fp:
        return JAML.load(fp)



from typing import Dict, Sequence, List, Optional, Any, Tuple

from . import Document, DocumentSourceType, typename, DocumentContentType
from ...excepts import BadDocType

__all__ = ['MultimodalDocument']


class MultimodalDocument(Document):
    """
    :class:`MultimodalDocument` is a data type created based on Jina primitive data type :class:`Document`.

    It shares the same methods and properties with :class:`Document`, while it focus on modality at chunk level.

    .. warning::
        - It assumes that every ``chunk`` of a ``document`` belongs to a different modality.
        - It assumes that every :class:`MultimodalDocument` have at least two chunks.
    """

    def __init__(self, document: Optional[DocumentSourceType] = None,
                 chunks: Sequence[Document] = None,
                 modality_content_map: Dict[str, DocumentContentType] = None,
                 copy: bool = False, **kwargs):
        """

        :param document: the document to construct from. If ``bytes`` is given
                then deserialize a :class:`DocumentProto`; ``dict`` is given then
                parse a :class:`DocumentProto` from it; ``str`` is given, then consider
                it as a JSON string and parse a :class:`DocumentProto` from it; finally,
                one can also give `DocumentProto` directly, then depending on the ``copy``,
                it builds a view or a copy from it.
        :param chunks: the chunks of the multimodal document to initialize with. Expected to
                received a list of :class:`Document`, with different modalities.
        :param: `modality_content_mapping`: A Python dict, the keys are the modalities and the values
                are the :attr:`content` of the :class:`Document`
        :param copy: when ``document`` is given as a :class:`DocumentProto` object, build a
                view (i.e. weak reference) from it or a deep copy from it.
        :param kwargs: other parameters to be set

        .. warning::
            - Build :class:`MultimodalDocument` from :attr:`modality_content_mapping` expects you assign
              :attr:`Document.content` as the value of the dictionary.
        """
        super().__init__(document=document, copy=copy, **kwargs)
        if chunks or modality_content_map:
            if chunks:
                granularities = [chunk.granularity for chunk in chunks]
                if len(set(granularities)) != 1:
                    raise BadDocType('Each chunk should have the same granularity.')
                self.chunks.extend(chunks)
            elif modality_content_map:
                self.modality_content_map = modality_content_map
            self._handle_chunk_level_attributes()

    @property
    def is_valid(self) -> bool:
        """A valid :class:`MultimodalDocument` should meet the following requirements:

            - Document should consist at least 2 chunks.
            - Length of modality is not identical to length of chunks.
        """
        modalities = set([chunk.modality for chunk in self.chunks])
        return 2 <= len(self.chunks) == len(modalities)

    def _handle_chunk_level_attributes(self):
        """Handle chunk attributes, such as :attr:`granularity` and :attr:`mime_type`.

        Chunk granularity should be greater than parent granularity level. Besides, if the chunk do not have
            a specified :attr:`mime_type`, it will be manually set to it's parent's :attr:`mime_type`.
        """

        # Joan: https://github.com/jina-ai/jina/pull/1335#discussion_r533905780
        # If chunk.granularity is 0. (This means a user without caring for granularity wants
        #   to merge N documents into a multimodal document, therefore we do what
        #   u have here of increasing their granularity inside this set) Well documented please
        # If the chunk comes with granularity > 0, then it means that someone has cared to chunk already
        #   the document or that we have some driver that generates muktimodal documents in the future.
        #   Then, have document.granularity = chunk.granularity - 1.
        for chunk in self.chunks:
            if chunk.granularity == 0:
                chunk.granularity = self.granularity + 1
            else:
                self.granularity = chunk.granularity - 1
            if not chunk.mime_type:
                chunk.mime_type = self.mime_type

    @property
    def modality_content_map(self) -> Dict:
        """Get the mapping of modality and content, the mapping is represented as a :attr:`dict`, the keys
        are the modalities of the chunks, the values are the corresponded content of the chunks.

        :return: the mapping of modality and content extracted from chunks.
        """
        result = {}
        for chunk in self.chunks:
            modality = chunk.modality
            result[modality] = chunk.embedding if chunk.embedding is not None else chunk.content
        return result

    @modality_content_map.setter
    def modality_content_map(self, value: Dict[str, Any]):
        for modality, content in value.items():
            with Document() as chunk:
                chunk.modality = modality
                chunk.content = content
                self.chunks.add(chunk)

    def __getitem__(self, modality: str) -> DocumentContentType:
        """Extract content by the name of the modality.

        :param modality: The name of the modality.
        :return: Content of the corresponded modality.
        """
        if isinstance(modality, str):
            return self.modality_content_map.get(modality, None)
        else:
            raise TypeError(f'{typename(modality)} is not supported')

    @property
    def modalities(self) -> List[str]:
        """Get all modalities of the :class:`MultimodalDocument`.

        :return: List of modalities extracted from chunks of the document.
        """
        return list(self.modality_content_map.keys())

    def update_content_hash(self, exclude_fields: Tuple[str] = ('id', 'matches', 'content_hash')) -> None:
        """ Update content hash of the document by including ``chunks`` when computing the hash
        """
        super().update_content_hash(exclude_fields)


from .base import set_base_parser


def set_check_parser(parser=None):
    if not parser:
        parser = set_base_parser()

    parser.add_argument('--summary-exec', type=str,
                        help='the markdown file path for all executors summary')
    parser.add_argument('--summary-driver', type=str,
                        help='the markdown file path for all drivers summary')
    return parser



from jina.parsers.helper import add_arg_group
from jina.enums import RemoteAccessType


def mixin_distributed_feature_parser(parser):
    """Mixing in arguments required by :class:`BasePod` into the given parser. """
    gp = add_arg_group(parser, title='Distributed')

    gp.add_argument('--remote-manager',
                    choices=list(RemoteAccessType),
                    default=RemoteAccessType.JINAD,
                    type=RemoteAccessType.from_string,
                    help=f'the manager of remote Jina')

    gp.add_argument('--silent-remote-logs', action='store_true', default=False,
                    help=f'do not display the streaming of remote logs on local console')

    gp.add_argument('--upload-files', type=str, nargs='*', metavar='FILE',
                    help='the files on the host to be uploaded to the remote workspace. This can be useful when your '
                         'Pod has more file dependencies beyond a single YAML file, e.g. Python files, data files. '
                         'Note (1) currently only flatten structure is supported, which means if you upload '
                         '`[./foo/a.py, ./foo/b.pp, ./bar/c.yml]`, then they will be put under the _same_ workspace on '
                         'the remote, losing all hierarchies. (2) By default, `--uses` YAML file is always uploaded. '
                         '(3) Uploaded files are by default isolated across the runs. To ensure files are submitted '
                         'to the same workspace across different runs, use `--workspace-id`.')

    gp.add_argument('--workspace-id', type=str,
                    help='the UUID for identifying the workspace on remote. When not given then remote will assign a '
                         'random one. Multiple Pea/Pod/Flow will work under the same workspace if they share the same '
                         '`workspace-id`.')



from ...helper import add_arg_group
from ....helper import random_port
from .... import __default_host__


def mixin_remote_parser(parser):
    gp = add_arg_group(parser, title='Expose')

    gp.add_argument('--host', type=str, default=__default_host__,
                    help=f'host address of the runtime, by default it is {__default_host__}.')

    gp.add_argument('--port-expose',
                    type=int,
                    default=random_port(),
                    help='host port exposed to the public')


def mixin_grpc_parser(parser=None):
    gp = add_arg_group(parser, title='GRPC/REST')

    gp.add_argument('--max-message-size', type=int, default=-1,
                    help='maximum send and receive size for gRPC server in bytes, -1 means unlimited')
    gp.add_argument('--proxy', action='store_true', default=False,
                    help='respect the http_proxy and https_proxy environment variables. '
                         'otherwise, it will unset these proxy variables before start. '
                         'gRPC seems to prefer no proxy')
    gp.add_argument('--prefetch', type=int, default=50,
                    help='the number of pre-fetched requests from the client')
    gp.add_argument('--prefetch-on-recv', type=int, default=1,
                    help='the number of additional requests to fetch on every receive')
    gp.add_argument('--restful', '--rest-api', action='store_true', default=False,
                    help='use RESTful interface instead of gRPC as the main interface')



import argparse

from .base import set_base_parser
from .helper import _SHOW_ALL_ARGS
from .peapods.base import mixin_base_ppr_parser
from ..enums import FlowOutputType, FlowOptimizeLevel, FlowInspectType


def set_flow_parser(parser=None):
    if not parser:
        parser = set_base_parser()

    mixin_base_ppr_parser(parser)

    parser.add_argument('--uses', type=str, help='a yaml file represents a flow')
    parser.add_argument('--inspect', type=FlowInspectType.from_string,
                        choices=list(FlowInspectType), default=FlowInspectType.COLLECT,
                        help='strategy on those inspect pods in the flow. '
                             'if REMOVE is given then all inspect pods are removed when building the flow')

    parser.add_argument('--optimize-level', type=FlowOptimizeLevel.from_string, default=FlowOptimizeLevel.NONE,
                        help='removing redundant routers from the flow. Note, this may change the gateway zmq socket to BIND \
                            and hence not allow multiple clients connected to the gateway at the same time.'
                        if _SHOW_ALL_ARGS else argparse.SUPPRESS)
    return parser



import os
import uuid
from contextlib import contextmanager
from pathlib import Path
from typing import Union

from daemon import jinad_args


def get_workspace_path(workspace_id: Union[uuid.UUID, str], *args):
    return os.path.join(jinad_args.workspace, str(workspace_id), *args)


@contextmanager
def jina_workspace(workspace_id: uuid.UUID):
    """
    Change the current working dir to ``path`` in a context and set it back to the original one when leaves the context.
    """
    old_dir = os.getcwd()
    old_var = os.environ.get('JINA_LOG_WORKSPACE', None)
    _workdir = get_workspace_path(workspace_id)
    Path(_workdir).mkdir(parents=True, exist_ok=True)
    os.environ['JINA_LOG_WORKSPACE'] = _workdir
    os.chdir(_workdir)
    try:
        yield _workdir
    finally:
        os.chdir(old_dir)
        if old_var:
            os.environ['JINA_LOG_WORKSPACE'] = old_var
        else:
            os.environ.pop('JINA_LOG_WORKSPACE')



from jina.executors.rankers import Chunk2DocRanker


class MaxRanker(Chunk2DocRanker):
    """
    :class:`MaxRanker` calculates the score of the matched doc from the matched chunks.
    For each matched doc, the score is the maximal score from all the matched chunks belonging to this doc.

    .. warning: Here we suppose that the larger chunk score means the more similar.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        import warnings
        warnings.warn("MaxRanker is deprecated. Please use SimpleAggregateRanker instead", DeprecationWarning,
                      stacklevel=2)

    def _get_score(self, match_idx, query_chunk_meta, match_chunk_meta, *args, **kwargs):
        return self.get_doc_id(match_idx), match_idx[self.COL_SCORE].max()



from ..helper import add_arg_group


def mixin_hub_list_parser(parser):
    gp = add_arg_group(parser, title='List')
    gp.add_argument('--name', type=str,
                    help='name of hub image')
    gp.add_argument('--kind', type=str,
                    help='kind of hub image')
    gp.add_argument('--keywords', type=str, nargs='+', metavar='KEYWORD',
                    help='keywords for searching')
    gp.add_argument('--type', type=str, default='pod', choices=['pod', 'app'],
                    help='type of the hub image')
    gp.add_argument('--local-only', action='store_true', default=False,
                    help='list all local hub images on the current machine')



import uuid

from fastapi import APIRouter, HTTPException

from jina.helper import ArgNamespace
from jina.parsers import set_pea_parser
from ... import Runtime400Exception
from ...models import PeaModel
from ...models.status import StoreStatus, StoreItemStatus
from ...stores import pea_store as store

router = APIRouter(prefix='/peas', tags=['peas'])


@router.get(
    path='',
    summary='Get all alive Pea\' status',
    response_model=StoreStatus
)
async def _get_items():
    return store.status


@router.get(
    path='/arguments',
    summary='Get all accept arguments of a Pea',
)
async def _fetch_pea_params():
    return PeaModel.schema()['properties']


@router.post(
    path='',
    summary='Create a Pea',
    description='Create a Pea and add it to the store',
    status_code=201,
    response_model=uuid.UUID
)
async def _create(pea: 'PeaModel'):
    try:
        args = ArgNamespace.kwargs2namespace(pea.dict(), set_pea_parser())
        return store.add(args)
    except Exception as ex:
        raise Runtime400Exception from ex


# order matters! this must be put in front of del {id}
#  https://fastapi.tiangolo.com/tutorial/path-params/?h=+path#order-matters
@router.delete(
    path='/all',
    summary='Terminate all running Peas',
)
async def _clear_all():
    store.clear()


@router.delete(
    path='/{id}',
    summary='Terminate a running Pea',
    description='Terminate a running Pea and release its resources'
)
async def _delete(id: uuid.UUID):
    try:
        del store[id]
    except KeyError:
        raise HTTPException(status_code=404, detail=f'{id} not found in {store!r}')


@router.get(
    path='/{id}',
    summary='Get status of a running Pea',
    response_model=StoreItemStatus
)
async def _status(id: uuid.UUID):
    try:
        return store[id]
    except KeyError:
        raise HTTPException(status_code=404, detail=f'{id} not found in {store!r}')


@router.on_event('shutdown')
def _shutdown():
    store.reset()



from .base import InputFnType, BaseClient, CallbackFnType
from .websockets import WebSocketClientMixin
from ..enums import RequestType


class AsyncClient(BaseClient):
    """
    :class:`AsyncClient` is the asynchronous version of the :class:`Client`. They share the same interface, except
    in :class:`AsyncClient` :meth:`train`, :meth:`index`, :meth:`search` methods are coroutines
    (i.e. declared with the async/await syntax), simply calling them will not schedule them to be executed.
    To actually run a coroutine, user need to put them in an eventloop, e.g. via ``asyncio.run()``,
    ``asyncio.create_task()``.

    :class:`AsyncClient` can be very useful in
    the integration settings, where Jina/Flow/Client is NOT the main logic, but rather served as a part of other program.
    In this case, users often do not want to let Jina control the ``asyncio.eventloop``. On contrary, :class:`Client`
    is controlling and wrapping the eventloop internally, making the Client looks synchronous from outside.

    For example, say you have the Flow running in remote. You want to use Client to connect to it do
    some index and search, but meanwhile you have some other IO-bounded jobs and want to do them concurrently.
    You can use :class:`AsyncClient`,

    .. highlight:: python
    .. code-block:: python

        from jina.clients.asyncio import AsyncClient

        ac = AsyncClient(...)

        async def jina_client_query():
            await ac.search(...)

        async def heavylifting():
            await other_library.download_big_files(...)

        async def concurrent_main():
            await asyncio.gather(jina_client_query(), heavylifting())


        if __name__ == '__main__':
            # under python
            asyncio.run(concurrent_main())

    One can think of :class:`Client` as Jina-managed eventloop, whereas :class:`AsyncClient` is self-managed eventloop.
    """

    async def train(self, input_fn: InputFnType = None,
                    on_done: CallbackFnType = None,
                    on_error: CallbackFnType = None,
                    on_always: CallbackFnType = None,
                    **kwargs) -> None:
        """

        :param input_fn: the input function that generates the content
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param kwargs:
        :return:
        """
        self.mode = RequestType.TRAIN
        return await self._get_results(input_fn, on_done, on_error, on_always, **kwargs)

    async def search(self, input_fn: InputFnType = None,
                     on_done: CallbackFnType = None,
                     on_error: CallbackFnType = None,
                     on_always: CallbackFnType = None,
                     **kwargs) -> None:
        """

        :param input_fn: the input function that generates the content
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param kwargs:
        :return:
        """
        self.mode = RequestType.SEARCH
        return await self._get_results(input_fn, on_done, on_error, on_always, **kwargs)

    async def index(self, input_fn: InputFnType = None,
                    on_done: CallbackFnType = None,
                    on_error: CallbackFnType = None,
                    on_always: CallbackFnType = None,
                    **kwargs) -> None:
        """

        :param input_fn: the input function that generates the content
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param kwargs:
        :return:
        """
        self.mode = RequestType.INDEX
        return await self._get_results(input_fn, on_done, on_error, on_always, **kwargs)


class AsyncWebSocketClient(AsyncClient, WebSocketClientMixin):
    """
    :class:`AsyncWebSocketClient` is the asynchronous version of the :class:`WebSocketClient`.
    They share the same interface, except in :class:`AsyncWebSocketClient` :meth:`train`, :meth:`index`, :meth:`search`
    methods are coroutines (i.e. declared with the async/await syntax), simply calling them will not schedule them to be executed.
    To actually run a coroutine, user need to put them in an eventloop, e.g. via ``asyncio.run()``,
    ``asyncio.create_task()``.

    :class:`AsyncWebSocketClient` can be very useful in the integration settings, where Jina/Flow/Client is NOT the
    main logic, but rather served as a part of other program. In this case, users often do not want to let Jina control
    the ``asyncio.eventloop``. On contrary, :class:`WebSocketClient` is controlling and wrapping the eventloop
    internally, making the Client looks synchronous from outside.

    For example, say you have the Flow running in remote. You want to use Client to connect to it do
    some index and search, but meanwhile you have some other IO-bounded jobs and want to do them concurrently.
    You can use :class:`AsyncWebSocketClient`,

    .. highlight:: python
    .. code-block:: python

        from jina.clients.asyncio import AsyncWebSocketClient

        ac = AsyncWebSocketClient(...)

        async def jina_client_query():
            await ac.search(...)

        async def heavylifting():
            await other_library.download_big_files(...)

        async def concurrent_main():
            await asyncio.gather(jina_client_query(), heavylifting())


        if __name__ == '__main__':
            # under python
            asyncio.run(concurrent_main())

    One can think of :class:`WebSocketClient` as Jina-managed eventloop,
    whereas :class:`AsyncWebSocketClient` is self-managed eventloop.
    """
